package org.apidesign.apifest08.currency;


import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.Currency;


/**
 * Convert between two currencies.
 *
 * @author D'Arcy Smith
 * @version 1.0
 */
final class ConvertorImpl
    implements Convertor
{
    /**
     * The currency to cvonvert from.
     */
    private final Currency currencyA;

    /**
     * The currency to cvonvert from.
     */
    private final Currency currencyB;

    /**
     * The echange rate between a and b.
     */
    private final BigDecimal currencyARate;

    /**
     * The echange rate between b and a.
     */
    private final BigDecimal currencyBRate;
    
    /**
     * Constructs a convertor with the specified currencies.
     * 
     * @param a the currency to convert from.
     * @param aRate the exchage rage between from and to.
     * @param b the currency to convert to.
     * @param bRate the exchage rage between to and from.
     * @throws IllegalArgumentException if either any of the arguments are null or if either rate <= 0.
     */
    public ConvertorImpl(final Currency   a,
                         final BigDecimal aRate,
                         final Currency   b,
                         final BigDecimal bRate)
    {
        if(a == null)
        {
            throw new IllegalArgumentException("a cannot be null");
        }

        if(b == null)
        {
            throw new IllegalArgumentException("b cannot be null");
        }

        if(aRate == null)
        {
            throw new IllegalArgumentException("aRate cannot be null");
        }

        if(bRate == null)
        {
            throw new IllegalArgumentException("bRate cannot be null");
        }
                
        if(aRate.compareTo(BigDecimal.ZERO) <= 0)
        {
            throw new IllegalArgumentException("aRate must be > 0, was: " + aRate);
        }
                
        if(bRate.compareTo(BigDecimal.ZERO) <= 0)
        {
            throw new IllegalArgumentException("bRate must be > 0, was: " + bRate);
        }
        
        currencyA     = a;
        currencyB     = b;
        currencyARate = aRate;
        currencyBRate = bRate;
    }
    
    /**
     * Convert an amount from one currency to another.
     * 
     * @param from the currency to convert from.
     * @param to the currency to convert to.
     * @param amount the amount to convert.
     * @return the converted amount.
     * @throws IllegalArgumentException if any of the arguments are null.
     * @throws InvalidConversionException if either from or to are not equal to the currencies passed to the constructor.
     */
    public BigDecimal convert(final Currency   from,
                              final Currency   to,
                              final BigDecimal amount)
        throws InvalidConversionException
    {
        final BigDecimal result;
        
        if(amount == null)
        {
            throw new IllegalArgumentException("amount cannot be null");
        }
        
        if(from == null)
        {
            throw new IllegalArgumentException("from cannot be null");
        }
        
        if(to == null)
        {
            throw new IllegalArgumentException("to cannot be null");
        }
        
        if(!(from.equals(currencyA)) && (!(from.equals(currencyB))))
        {
            throw new InvalidConversionException("cannot convert from: " + from.getCurrencyCode(), from, currencyA, currencyB);
        }
        
        if(!(to.equals(currencyA)) && (!(to.equals(currencyB))))
        {
            throw new InvalidConversionException("cannot convert to: " + to.getCurrencyCode(), to, currencyA, currencyB);
        }

        // converting between the same currency is no converstion at all.
        if(from.equals(to))
        {
            result = amount;
        }
        else
        {
            final BigDecimal rateX;
            final BigDecimal rateY;
            final BigDecimal temp;
            
            if(from.equals(currencyA))
            {
                rateX = currencyARate;
                rateY = currencyBRate;
            }
            else
            {
                rateX = currencyBRate;
                rateY = currencyARate;
            }

            temp   = amount.divide(rateX, MathContext.DECIMAL32);
            result = temp.multiply(rateY);
        }

        return (result.setScale(2, RoundingMode.HALF_DOWN));
    }
}
